package com.zhengyiyu.feature;

import com.zhengyiyu.util.SequenceUtil;

/**
 * @author yiyu
 * all the DNA composition features used in prediction
 * now it is only for usage in nucleosome position prediction
 */
public class CompositionFeature extends SequenceFeature {
	/**
	 * The length of the pattern, if the input sequence is shorter than the pattern length, we cannot handle it.
	 */
	private int patternLength;
	
	/**
	 * The main pattern of this features, cannot be null
	 */
	private String mainPattern;
	
	/**
	 * Usually it is the complementary reversed sequence of the main pattern, for example, AAAA - TTTT, AC - GT, or null if the complementary reversed sequence is same or just no subpattern
	 */
	private String secondPattern;
	
	
	/**
	 * @param name
	 *            the name is just the pattern, if the name contains "/" means
	 *            that the sub pattern is included, e.g., A/T
	 */
	public CompositionFeature(String name) {
		super();
		this.setFeatureName(name.replace("/", "_"));
		
		if (name.contains("/")) {
			String[] result = name.split("/");
			this.patternLength = result[0].length();
			this.mainPattern = result[0];
			this.secondPattern = result[1];
			
			if (!SequenceUtil.getReverseSequence(SequenceUtil.getComplementSequence(mainPattern)).equals(secondPattern)) {
				System.out.println("No Way!:" + name);
			}
		} else if (name.contains("_")) {
			String[] result = name.split("_");
			this.patternLength = result[0].length();
			this.mainPattern = result[0];
			this.secondPattern = result[1];
			
			if (!SequenceUtil.getReverseSequence(SequenceUtil.getComplementSequence(mainPattern)).equals(secondPattern)) {
				System.out.println("No Way!:" + name);
			}
		} else {
			this.patternLength = name.length();
			this.mainPattern = name;
			this.secondPattern = "";
		}
	}
	
	/**
	 * @param name
	 *            information about the pattern, usually like "AAAA/TTTT" which
	 *            can be interpretated as pattern AAAA, secondPattern TTTT
	 * @param reference
	 *            the reference info of the pattern
	 */
	public CompositionFeature(String name, String reference) {
		this(name);
		this.setReference(reference);
	}
	
	/**
	 * @param name
	 * @param patternName
	 *            label used in patterns, like "A"
	 * @param cutoffs
	 *            discretization cutoffs for the features
	 */
	public CompositionFeature(String name, String patternName, double[] cutoffs) {
		this(name);
		this.setPatternName(patternName);
		
		for (double d : cutoffs) {
			this.getDiscretizationCutoff().add(d);
		}
	}
	
	/**
	 * @param name
	 *            pattern name
	 * @param length
	 *            pattern length
	 * @param pattern
	 *            main pattern
	 * @param subPattern
	 *            second pattern, usually the complementary reversed pattern
	 */
	public CompositionFeature(String name, int length, String pattern, String subPattern) {
		super();
		this.setFeatureName(name.replace("/", "_"));
		this.patternLength = length;
		this.mainPattern = pattern;
		this.secondPattern = subPattern;
	}
	
	/**
	 * @return the length
	 */
	public int getPatternLength() {
		return patternLength;
	}
	
	/**
	 * @param length the length to set
	 */
	public void setPatternLength(int length) {
		this.patternLength = length;
	}
	
	/**
	 * @return the pattern
	 */
	public String getPattern() {
		return mainPattern;
	}
	
	/**
	 * @param pattern the pattern to set
	 */
	public void setPattern(String pattern) {
		this.mainPattern = pattern;
	}
	
	/**
	 * @return the subPattern
	 */
	public String getSubPattern() {
		return secondPattern;
	}
	
	/**
	 * @param subPattern the subPattern to set
	 */
	public void setSubPattern(String subPattern) {
		this.secondPattern = subPattern;
	}

	/**
	 * @param seq
	 * @return the double value of all the seq
	 * 
	 */
	public double getFeature(String seq) {
		return this.calculateFeatureValue(seq);
	}
	
	/**
	 * @param seq
	 * @param length
	 * @return the feature value of the central subseq whose length is as length
	 */
	public double getFeature(String seq, int length) {
		if (length < 0) {
			return getFeature(seq);
		}
		if (seq.length() < length) {
			return getFeature(seq);
		} else {
			int startIndex = (seq.length() - length) / 2;
			return getFeature(seq.substring(startIndex, startIndex + length));
		}
	}
	
	/* (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		return this.getFeatureName().hashCode();
	}

	/* (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		
		CompositionFeature other = (CompositionFeature) obj;
		if (this.getFeatureName() == null) {
			if (other.getFeatureName() != null) {
				return false;
			}
		} else if (!this.getFeatureName().equals(other.getFeatureName())) {
			return false;
		}
		
		return true;
	}

	
	@Override
	public double calculateFeatureValue(String seq) {
		double result = countFeature(seq);
		return result * 1.0 / (seq.length() - this.patternLength + 1);
	}

	public double countFeature(String seq) {
		if (seq == null || seq.equals("")) {
			System.out.println("Composition Feature error: Sequence cannot be null to calculate feature values");
			return -Double.MAX_VALUE;
		}
		
		if (seq.length() < this.patternLength) {
			System.out.println("Composition Feature error: the Length is less than the pattern Length");
			return -Double.MAX_VALUE;
		}
		
		double result = 0;
		
		String sub = "";
		for (int i = 0; i <= seq.length() - this.patternLength; i++) {
			sub = seq.substring(i, i + this.patternLength);
			if (sub.equals(this.mainPattern) || sub.equals(this.secondPattern)) {
				result++;
			}
		}
		return result;
	}
	
	/**
	 * @param seq
	 * @return the log value of the feature value
	 */
	public double getLogFeature(String seq) {
		double featureValue = getFeature(seq);
		return Math.log(featureValue);
	}

	/**
	 * @param seq
	 * @param length
	 * @return the log value of the feature value
	 */
	public double getLogFeature(String seq, int length) {
		double average = getFeature(seq, length);
		if (average == 0.0) {
			return 0;
		}
		return Math.log(average);
	}
	
}
